// arduino.js

(function (ext) {
  var COMMAND = {
    PULLUP_SWITCH: 100,
    LCD_PRINT: 101,
    LCD_CLEAR: 102
  }

  var PIN = {
    POTENTION: 'A0',
    MIC: 'A1',
    TACT_1: 'A2',
    TACT_2: 'A3',
    LCD_I2C_SDA: 'A4',
    LCD_I2C_SCL: 'A5',
    BUZZER: 2,
    LINE: 3,
    ULTRASONIC_TRIG: 4,
    ULTRASONIC_ECHO: 7,
    SERVO: 5,
    DCM_A: 6,
    DCM_B: 9,
    LED1: 10,
    LED2: 11,
    RGB_R: 8,
    RGB_G: 13,
    RGB_B: 12
  };
  var device = null;
  var _rxBuf = [];

  // Sensor states:
  var ports = {
    Port1: 1,
    Port2: 2,
    Port3: 3,
    Port4: 4,
    Port5: 5,
    Port6: 6,
    Port7: 7,
    Port8: 8,
    M1: 9,
    M2: 10
  };
  var slots = {
    Slot1: 1,
    Slot2: 2
  };
  var switchStatus = {
    On: 1,
    Off: 0
  };
  var levels = {
    HIGH: 1,
    LOW: 0
  };
  var DCM_DIRECTION = {
    forward: 1,
    reverse: 0
  };
  var LED_LEVEL = {
    "led_on": 255,
    "led_off": 0
  };
  var LED_PIN_MAP = {
    LED1: 10,
    LED2: 11
  };
  var SWITCH_PIN_MAP = {
    switch1: 2,
    switch2: 3
  };
  var axis = {
    'X-Axis': 1,
    'Y-Axis': 2,
    'Z-Axis': 3
  }
  var tones = {
    "B0": 31, "C1": 33, "D1": 37, "E1": 41, "F1": 44, "G1": 49, "A1": 55, "B1": 62,
    "C2": 65, "D2": 73, "E2": 82, "F2": 87, "G2": 98, "A2": 110, "B2": 123,
    "C3": 131, "D3": 147, "E3": 165, "F3": 175, "G3": 196, "A3": 220, "B3": 247,
    "C4": 262, "D4": 294, "E4": 330, "F4": 349, "G4": 392, "A4": 440, "B4": 494,
    "C5": 523, "D5": 587, "E5": 659, "F5": 698, "G5": 784, "A5": 880, "B5": 988,
    "C6": 1047, "D6": 1175, "E6": 1319, "F6": 1397, "G6": 1568, "A6": 1760, "B6": 1976,
    "C7": 2093, "D7": 2349, "E7": 2637, "F7": 2794, "G7": 3136, "A7": 3520, "B7": 3951,
    "C8": 4186, "D8": 4699
  };
  var beats = {"Half": 500, "Quarter": 250, "Eighth": 125, "Whole": 1000, "Double": 2000, "Zero": 0};
  var values = {};
  var indexs = [];

  var startTimer = 0;
  var versionIndex = 0xFA;
  ext.resetAll = function () {
    device.send([ 0xff, 0x55, 2, 0, 4 ]);
  };

  ext.runArduino = function () {
    responseValue();
  };
  ext.runDigital = function (pin, level) {
    runPackage(30, pin, typeof level == "string" ? levels[ level ] : new Number(level));
  };
  ext.writeLED = function (pin, level) {
    var pinValue = typeof pin == "string" ? LED_PIN_MAP[ pin ] : new Number(pin);
    var levelValue = typeof level == "string" ? LED_LEVEL[ level ] : new Number(level);

    runPackage(32, pinValue, levelValue);
  };
  ext.writeRGB = function (R, G, B) {
    runPackage(30, PIN.RGB_R, typeof R == "string" ? levels[ R ] : new Number(R));
    runPackage(30, PIN.RGB_G, typeof G == "string" ? levels[ G ] : new Number(G));
    runPackage(30, PIN.RGB_B, typeof B == "string" ? levels[ B ] : new Number(B));
  };
  ext.runPwm = function (pin, pwm) {
    runPackage(32, pin, pwm);
  };
  ext.runDCM = function (speed, direction) {
    var dcmASpeed = 0;
    var dcmBSpeed = 0;

    switch(speed){
      case "max":
        speed = 255;
        break;
      case "min":
        speed = 0;
        break;
      default:
        speed = speed;
        break;
    }

    speed = Math.round(speed);
    speed = Math.max(speed, 0);
    speed = Math.min(speed, 255);


    if (DCM_DIRECTION[ direction ] == 1) {
      dcmASpeed = speed;
      dcmBSpeed = 0;
    } else {
      dcmASpeed = 0;
      dcmBSpeed = speed;
    }
    runPackage(32, PIN.DCM_A, dcmASpeed);
    runPackage(32, PIN.DCM_B, dcmBSpeed);
  };
  ext.runTone = function (tone, beat) {
    runPackage(34, PIN.BUZZER, short2array(typeof tone == "number" ? tone : tones[ tone ]), short2array(typeof beat == "number" ? beat : beats[ beat ]));
  };
  ext.runServo = function (angle) {
    runPackage(33, PIN.SERVO, angle);
  };
  ext.writeLCD = function (row, column, str) {
    var chars = [];
    var wordArray = String(str).split('');

    for (var i = 0; i < wordArray.length; i++) {
      var char = wordArray[ i ].charCodeAt(0);

      chars.push(char);
    }

//      아래코드에서 보조 변수들(script.isStart, script.timeFlag등)이 들어간 이유는 fps(초당프레임)를 위해서입니다.
//      해당 코드가 없을 경우 최소 딜레이가 없기때문에 흐리게 나오는 문제가 발생합니다.
//    if (!script.isStart) {
//      if (!sq[ 'SET' ]) {
//        sq[ 'SET' ] = {};
//      }
//
//      script.isStart = true;
//      script.timeFlag = 1;
//      var fps = Entry.FPS || 60;
//      var timeValue = 60 / fps * 50;
//
//      runPackage(COMMAND.LCD_PRINT, row - 1, column - 1, chars.length, chars);
//
//      setTimeout(function () {
//        script.timeFlag = 0;
//      }, timeValue);
//      return script;
//    } else if (script.timeFlag == 1) {
//      return script;
//    } else {
//      delete script.timeFlag;
//      delete script.isStart;
//      Entry.engine.isContinue = false;
//      runPackage(COMMAND.LCD_PRINT, row - 1, column - 1, chars.length, chars);
//    }

    runPackage(COMMAND.LCD_PRINT, row - 1, column - 1, chars.length, chars);
  };
  ext.clearLCD = function () {
    runPackage(COMMAND.LCD_CLEAR);
  };
  ext.readLineSensor = function (nextID) {
    var deviceId = 30;
    getPackage(nextID, deviceId, PIN.LINE);
  };
  ext.readSwitch = function (nextID, pin) {
    getPackage(nextID, COMMAND.PULLUP_SWITCH, SWITCH_PIN_MAP[ pin ]);
  };
  ext.readMic = function (nextID) {
    var deviceId = 31;
    getPackage(nextID, deviceId, 1);
  };
  ext.readPotentiometer = function (nextID) {
    var deviceId = 31;
    getPackage(nextID, deviceId, 0);
  };
  ext.getUltrasonicArduino = function (nextID, trig, echo) {
    var deviceId = 36;
    getPackage(nextID, deviceId, PIN.ULTRASONIC_TRIG, PIN.ULTRASONIC_ECHO);
  }

  ext.resetTimer = function () {
    startTimer = new Date().getTime();
    responseValue();
  };
  ext.getPulse = function (nextID, pin, timeout) {
    var deviceId = 37;
    getPackage(nextID, deviceId, pin, short2array(timeout));
  };
  ext.getTimer = function (nextID) {
    if (startTimer == 0) {
      startTimer = new Date().getTime();
    }
    responseValue(nextID, new Date().getTime() - startTimer);
  }

  function sendPackage(argList, type) {
    var bytes = [ 0xff, 0x55, 0, 0, type ];

    for (var i = 0; i < argList.length; ++i) {
      var val = argList[ i ];
      if (Array.isArray(val)) {
        bytes = bytes.concat(val);
      } else {
        bytes.push(val);
      }
    }
    bytes[ 2 ] = bytes.length - 3;
    device.send(bytes);
  }

  function runPackage() {
    sendPackage(arguments, 2);
  }

  function getPackage() {
    var nextID = arguments[ 0 ];
    Array.prototype.shift.call(arguments);
    sendPackage(arguments, 1);
  }

  var inputArray = [];
  var _isParseStart = false;
  var _isParseStartIndex = 0;

  function processData(bytes) {
    var len = bytes.length;
    if (_rxBuf.length > 30) {
      _rxBuf = [];
    }
    for (var index = 0; index < bytes.length; index++) {
      var c = bytes[ index ];
      _rxBuf.push(c);
      if (_rxBuf.length >= 2) {
        if (_rxBuf[ _rxBuf.length - 1 ] == 0x55 && _rxBuf[ _rxBuf.length - 2 ] == 0xff) {
          _isParseStart = true;
          _isParseStartIndex = _rxBuf.length - 2;
        }
        if (_rxBuf[ _rxBuf.length - 1 ] == 0xa && _rxBuf[ _rxBuf.length - 2 ] == 0xd && _isParseStart) {
          _isParseStart = false;

          var position = _isParseStartIndex + 2;
          var extId = _rxBuf[ position ];
          position++;
          var type = _rxBuf[ position ];
          position++;
          //1 byte 2 float 3 short 4 len+string 5 double
          var value;
          switch (type) {
            case 1: {
              value = _rxBuf[ position ];
              position++;
            }
              break;
            case 2: {
              value = readFloat(_rxBuf, position);
              position += 4;
              if (value < -255 || value > 1023) {
                value = 0;
              }
            }
              break;
            case 3: {
              value = readInt(_rxBuf, position, 2);
              position += 2;
            }
              break;
            case 4: {
              var l = _rxBuf[ position ];
              position++;
              value = readString(_rxBuf, position, l);
            }
              break;
            case 5: {
              value = readDouble(_rxBuf, position);
              position += 4;
            }
              break;
            case 6:
              value = readInt(_rxBuf, position, 4);
              position += 4;
              break;
          }
          if (type <= 6) {
            responseValue(extId, value);
          } else {
            responseValue();
          }
          _rxBuf = [];
        }
      }
    }
  }

  function readFloat(arr, position) {
    var f = [ arr[ position ], arr[ position + 1 ], arr[ position + 2 ], arr[ position + 3 ] ];
    return parseFloat(f);
  }

  function readInt(arr, position, count) {
    var result = 0;
    for (var i = 0; i < count; ++i) {
      result |= arr[ position + i ] << (i << 3);
    }
    return result;
  }

  function readDouble(arr, position) {
    return readFloat(arr, position);
  }

  function readString(arr, position, len) {
    var value = "";
    for (var ii = 0; ii < len; ii++) {
      value += String.fromCharCode(_rxBuf[ ii + position ]);
    }
    return value;
  }

  function appendBuffer(buffer1, buffer2) {
    return buffer1.concat(buffer2);
  }

  // Extension API interactions
  var potentialDevices = [];
  ext._deviceConnected = function (dev) {
    potentialDevices.push(dev);

    if (!device) {
      tryNextDevice();
    }
  }

  function tryNextDevice() {
    // If potentialDevices is empty, device will be undefined.
    // That will get us back here next time a device is connected.
    device = potentialDevices.shift();
    if (device) {
      device.open({stopBits: 0, bitRate: 115200, ctsFlowControl: 0}, deviceOpened);
    }
  }

  var watchdog = null;

  function deviceOpened(dev) {
    if (!dev) {
      // Opening the port failed.
      tryNextDevice();
      return;
    }
    device.set_receive_handler('arduino', processData);
  };

  ext._deviceRemoved = function (dev) {
    if (device != dev) return;
    device = null;
  };

  ext._shutdown = function () {
    if (device) device.close();
    device = null;
  };

  ext._getStatus = function () {
    if (!device) return {status: 1, msg: 'Arduino disconnected'};
    if (watchdog) return {status: 1, msg: 'Probing for Arduino'};
    return {status: 2, msg: 'Arduino connected'};
  }

  var descriptor = {};
  ScratchExtensions.register('Arduino', descriptor, ext, {type: 'serial'});
})({});
